/*
    liblihata - list/hash/table format, utils
    Copyright (C) 2013  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project URLs: http://repo.hu/projects/lihata
                  svn://repo.hu/lihata


   This file implements a context wrapper around the event parser
   and does document validation with remembering minimal amount of
   data.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "liblihata/parser.h"
#include "liblihata/hash_str.h"
#include "liblihata/genht/htsp.h"
#include "lib_validator.h"

typedef struct {
	lht_node_type_t nt;
	union {
		htsp_t *hash;  /* all node names for a hash */
		int cols;      /* number of columns for a table */
		int items;     /* number of it items on a list */
	} data;
} lhtv_stack_t;

typedef struct lht_validate_s {
	int sused, salloced;
	lhtv_stack_t *stack;
	lht_parse_t orig_ctx;
} lht_validate_t;
#define STACK_GROWTH 16

/* returns previous top - valid until the next call to push() */
static lhtv_stack_t *pop(lht_validate_t *v)
{
	v->sused--;
	assert(v->sused >= 0);
	return v->stack + v->sused;
}

static lhtv_stack_t *top(lht_validate_t *v)
{
	assert(v->sused > 0);
	return v->stack + v->sused - 1;
}

/* returns new top - valid until the it's popped (... and then the next push()) */
static lhtv_stack_t *push(lht_validate_t *v, lht_node_type_t nt)
{
	lhtv_stack_t *ret;
	if (v->sused >= v->salloced) {
		v->salloced += STACK_GROWTH;
		v->stack = realloc(v->stack, v->salloced * sizeof(lhtv_stack_t));
	}
	ret = v->stack + v->sused;
	ret->nt = nt;
	memset(&(ret->data), 0, sizeof(ret->data));
	v->sused++;
	return ret;
}

static void lht_validate_event(lht_parse_t *ctx, lht_event_t ev, lht_node_type_t nt, const char *name, const char *value)
{
	lhtv_stack_t *sp, *parent;
	lht_validate_t *v = ctx->user_data;
	lht_err_t err = LHTE_SUCCESS;

	if (v->sused > 0) {
		sp = top(v);
		/* validate per node type */
		switch(sp->nt) {
			case LHT_LIST:
				/* count items for table */
				if ((ev == LHT_OPEN) || (ev == LHT_TEXTDATA))
					sp->data.items++;
				break;
			case LHT_HASH:
				/* check for duplicate entries */
				if ((ev == LHT_OPEN) || (ev == LHT_TEXTDATA)) {
					if (htsp_getentry(sp->data.hash, (char *)name) != NULL)
						err = LHTE_DUPLICATE_KEY;
					htsp_set(sp->data.hash, strdup(name), NULL);
				}
				break;

			default:;
		}
	}

	switch(ev) {
		case LHT_OPEN:
			sp = push(v, nt);
			if (sp->nt == LHT_HASH)
				sp->data.hash = htsp_alloc(lht_str_keyhash, lht_str_keyeq);
			else if (sp->nt == LHT_TABLE)
				sp->data.cols = -1;
			break;
		case LHT_CLOSE:
			sp = pop(v);
#warning TODO: free all strdup()d keys
			if (sp->nt == LHT_HASH)
				htsp_free(sp->data.hash);
			if (v->sused > 0) {
				parent = top(v);
				if (parent->nt == LHT_TABLE) {
					/* parent is a table, sp is the node last closed under table */
					if (sp->nt == LHT_LIST) {
						if (parent->data.cols >= 0) { /* not the first column, validate length */
							if (sp->data.items != parent->data.cols)
								err = LHTE_TABLE_COLS;
						}
						else /* first column, store length */
							parent->data.cols = sp->data.items;
					}
					else
						err = LHTE_TABLE_ROW_TYPE;
					
				}
			}
			break;
		case LHT_TEXTDATA:
			if (sp->nt == LHT_TABLE)
				err = LHTE_TABLE_ROW_TYPE;
			break;
		case LHT_COMMENT:
		case LHT_EOF:
		case LHT_ERROR:
			break;
	}

	/* make sure the fake ctx looks like the real one */
	v->orig_ctx.line = ctx->line;
	v->orig_ctx.col  = ctx->col;

	/* pass on the event */
	v->orig_ctx.event(&(v->orig_ctx), ev, nt, name, value);

	/* if we also had an error, pass on the error as well */
	if (err != LHTE_SUCCESS)
		v->orig_ctx.event(&(v->orig_ctx), LHT_ERROR, LHT_INVALID_TYPE, lht_err_str(err), NULL);
}

lht_parse_t *lht_validator_init(void *user_data, void (*event)(lht_parse_t *ctx, lht_event_t ev, lht_node_type_t nt, const char *name, const char *value))
{
	lht_parse_t *p;
	lht_validate_t *v;

	v = calloc(sizeof(lht_validate_t), 1);
	p = malloc(sizeof(lht_parse_t));
	v->orig_ctx.event = event;
	v->orig_ctx.user_data = user_data;
	lht_parser_init(p);
	p->event = lht_validate_event;
	p->user_data = v;
	return p;
}

void lht_validator_uninit(lht_parse_t *ctx)
{
	lht_validate_t *v = ctx->user_data;

	lht_parser_uninit(ctx);

	while(v->sused > 0) {
		lhtv_stack_t *sp;
		sp = pop(v);
		if (sp->nt == LHT_HASH)
			htsp_free(sp->data.hash);
	}
	free(v->stack);
	free(v);
	free(ctx);
}

